import json
import http3


async def broadcast_tx(blockstream_url, network, tx_hex):
    """Broadcast transaction
    https://github.com/Blockstream/esplora/blob/master/API.md#post-tx

    :param blockstream_url: blockstream url
    :type blockstream_url: str
    :param network: testnet or mainnet
    :type network: str
    :param tx_hex: tranaction hex
    :type tx_hex: str
    :returns: Transaction ID
    """
    try:
        if network == 'testnet':
            api_url = f'{blockstream_url}/testnet/api/tx'
        else:
            api_url = f'{blockstream_url}/api/tx'

        client = http3.AsyncClient(timeout=5)
        response = await client.post(url=api_url, data=tx_hex)

        if response.status_code == 200:
            res = str(response.content.decode('utf-8'))
            return res
        else:
            raise Exception(str(response.content.decode('utf-8')))
    except Exception as err:
        raise Exception(str(err))


import asyncio
import json
import http3
from . models.common import UTXO, Witness_UTXO
from . sochain_api import get_is_tx_confirmed
from xchainpy_client.models.balance import Balance
from xchainpy_util.asset import AssetBTC

HASKOIN_API_URL = 'https://api.haskoin.com/btc'

def haskoin_utxo_to_xchain_utxo(utxo):
    """Get utxo object from a sochain utxo

    :param utxo: haskoin utxo
    :type utxo: dict
    :returns: UTXO object
    """
    hash = utxo['txid']
    index = utxo['index']
    value = round(float(utxo['value']) * 10 ** 8)
    script =  bytearray.fromhex(utxo['pkscript']) #utxo['script_hex']
    witness_utxo = Witness_UTXO(value, script)
    return UTXO(hash, index, witness_utxo)

async def get_balance(address:str):
    """Get address balance
    https://api.haskoin.com/#/Address/getBalance

    :param address: wallet address
    :type address: str
    :returns: BTC balance
    """
    try:
        api_url = f'{HASKOIN_API_URL}/address/{address}/balance'

        client = http3.AsyncClient(timeout=5)
        response = await client.get(api_url)

        if response.status_code == 200:
            balance_response = json.loads(response.content.decode('utf-8'))
            confirmed = float(balance_response['confirmed'])
            unconfirmed = float(balance_response['unconfirmed'])
            total = confirmed + unconfirmed
            balance = [Balance(asset=AssetBTC, amount=total)]
            return balance
        else:
            return None
    except Exception as err:
        raise Exception(str(err))

async def get_unspent_txs(address, offset=0):
    """Get address balance
    https://api.haskoin.com/#/Address/getAddressUnspent

    :param address: address
    :type address: str
    :returns: A list of utxo's
    """
    try:
        api_url = f'{HASKOIN_API_URL}/address/{address}/unspent'

        if offset:
            api_url += f'?offset={offset}'

        client = http3.AsyncClient(timeout=5)
        response = await client.get(api_url)

        if response.status_code == 200:
            txs = json.loads(response.content.decode('utf-8'))
            if len(txs) == 100:
                # fetch the next batch
                offset += 100
                next_batch = await get_unspent_txs(address, offset)
                txs = txs + next_batch
            return txs

    except Exception as err:
        raise Exception(str(err))

async def get_confirmed_unspent_txs(address):
    """Get address balance
    https://api.haskoin.com/#/Address/getAddressUnspent

    :param address: address
    :type address: str
    :returns: A list confirmed of utxo's
    """
    txs = await get_unspent_txs(address)
    confirmed_UTXOs = await asyncio.wait([get_is_tx_confirmed(tx, True) for tx in txs])

    return confirmed_UTXOs


class DerivePath:
    def __init__(self, index:int=0):

        self._mainnet = f"84'/0'/0'/0/{index}"
        self._testnet = f"84'/1'/0'/0/{index}"

    @property
    def mainnet(self):
        return self._mainnet

    @mainnet.setter
    def mainnet(self, mainnet):
        self._mainnet = mainnet

    @property
    def testnet(self):
        return self._testnet

    @testnet.setter
    def testnet(self, testnet):
        self._testnet = testnet


class UTXO:
    def __init__(self, hash : str , index : int , witness_utxo , tx_hex:str=None):
      self._hash = hash
      self._index = index
      self._witness_utxo = witness_utxo
      self._tx_hex = tx_hex
    
    @property
    def hash(self):
        return self._hash

    @hash.setter
    def hash(self , hash):
        self._hash = hash
    
    @property
    def index(self):
        return self._index

    @index.setter
    def index(self , index):
        self._index = index

    @property
    def witness_utxo(self):
        return self._witness_utxo

    @witness_utxo.setter
    def witness_utxo(self , witness_utxo):
        self._witness_utxo = witness_utxo

    @property
    def tx_hex(self):
        return self._tx_hex

    @tx_hex.setter
    def tx_hex(self , tx_hex):
        self._tx_hex = tx_hex


class Witness_UTXO:
    def __init__(self, value, script):
      self._value = value
      self._script = script
    
    @property
    def value(self):
        return self._value

    @value.setter
    def value(self , value):
        self._value = value
    
    @property
    def script(self):
        return self._script

    @script.setter
    def script(self , script):
        self._script = script


import json
from . models.common import UTXO, Witness_UTXO
import http3
import asyncio
from xchainpy_client.models.balance import Balance
from xchainpy_util.asset import AssetBTC

DEFAULT_SUGGESTED_TRANSACTION_FEE = 127

def to_sochain_network(network: str):
    return 'BTCTEST' if network == 'testnet' else 'BTC'

def sochain_utxo_to_xchain_utxo(utxo):
    """Get utxo object from a sochain utxo

    :param utxo: sochain utxo
    :type utxo: dict
    :returns: UTXO object
    """
    hash = utxo['txid']
    index = utxo['output_no']
    
    value = round(float(utxo['value']) * 10 ** 8)
    script =  bytearray.fromhex(utxo['script_hex']) #utxo['script_hex']
    witness_utxo = Witness_UTXO(value, script)
    return UTXO(hash, index, witness_utxo)


async def get_transactions(sochain_url:str, network:str, address:str):
    """Get address information
    https://sochain.com/api#get-display-data-address

    :param sochain_url: sochain url
    :type sochain_url: str
    :param net: mainnet or testnet
    :type net: str
    :param address: wallet address
    :type address: str
    :returns: The fees with memo
    """
    try:
        api_url = f'{sochain_url}/address/{to_sochain_network(network)}/{address}'

        client = http3.AsyncClient(timeout=5)
        response = await client.get(api_url)

        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'))['data']
        else:
            return None
    except Exception as err:
        raise Exception(str(err))


async def get_tx(sochain_url:str, network:str, hash:str):
    """Get transaction by hash
    https://sochain.com/api#get-tx

    :param sochain_url: sochain url
    :type sochain_url: str
    :param net: mainnet or testnet
    :type net: str
    :param hash: The transaction hash
    :type hash: str
    :returns: The fees with memo
    """
    try:
        api_url = f'{sochain_url}/get_tx/{to_sochain_network(network)}/{hash}'

        client = http3.AsyncClient(timeout=5)
        response = await client.get(api_url)

        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'))['data']
        else:
            return None
    except Exception as err:
        raise Exception(str(err))


async def get_suggested_tx_fee():
    """Get Bitcoin suggested transaction fee
    Note: sochain does not provide fee rate related data
    Refer: https://app.bitgo.com/api/v2/btc/tx/fee

    :returns: The Bitcoin suggested transaction fee per bytes in sat
    """
    try:
        api_url = 'https://app.bitgo.com/api/v2/btc/tx/fee'

        client = http3.AsyncClient(timeout=5)
        response = await client.get(api_url)

        if response.status_code == 200:
            response = json.loads(response.content.decode('utf-8'))
            return response['feePerKb'] / 1000
        else:
            return DEFAULT_SUGGESTED_TRANSACTION_FEE ###
    except Exception as err:
        raise Exception(str(err))


async def get_balance(sochain_url:str, network:str, address:str):
    """Get address balance
    https://sochain.com/api#get-balance

    :param sochain_url: sochain url
    :type sochain_url: str
    :param network: mainnet or testnet
    :type network: str
    :param address: wallet address
    :type address: str
    :returns: BTC balance
    """
    try:
        api_url = f'{sochain_url}/get_address_balance/{to_sochain_network(network)}/{address}'

        client = http3.AsyncClient(timeout=5)
        response = await client.get(api_url)

        if response.status_code == 200:
            balance_response = json.loads(response.content.decode('utf-8'))['data']
            confirmed = float(balance_response['confirmed_balance'])
            unconfirmed = float(balance_response['unconfirmed_balance'])
            total = confirmed + unconfirmed
            balance = [Balance(asset=AssetBTC, amount=total)]
            return balance
        else:
            return None
    except Exception as err:
        raise Exception(str(err))

async def get_unspent_txs(sochain_url, network, address, starting_from_tx_id=None):
    """Get Unspent transactions
    https://sochain.com/api#get-unspent-tx

    :param sochain_url: sochain url
    :type sochain_url: str
    :param network: testnet or mainnet
    :type network: str
    :param address: address
    :type address: str
    :param starting_from_tx_id: starting_from_tx_id
    :type starting_from_tx_id: str
    :returns: A list of utxo's
    """
    try:
        api_url = f'{sochain_url}/get_tx_unspent/{to_sochain_network(network)}/{address}'

        if starting_from_tx_id:
            api_url += f'/{starting_from_tx_id}'

        client = http3.AsyncClient(timeout=5)
        response = await client.get(api_url)

        if response.status_code == 200:
            txs = json.loads(response.content.decode('utf-8'))['data']['txs']
            if len(txs) == 100:
                # fetch the next batch
                last_tx_id = txs[99]['txid']
                next_batch = await get_unspent_txs(sochain_url, network, address, last_tx_id)
                txs = txs + next_batch
            return txs
    except Exception as err:
        raise Exception(str(err))

async def get_confirmed_unspent_txs(sochain_url, network, address):
    """Get confirmed Unspent transactions
    https://sochain.com/api#get-unspent-tx

    :param sochain_url: sochain url
    :type sochain_url: str
    :param network: testnet or mainnet
    :type network: str
    :param address: address
    :type address: str
    :returns: A list confirmed of utxo's
    """
    txs = await get_unspent_txs(sochain_url, network, address)
    confirmed_UTXOs = await asyncio.gather(*[get_is_tx_confirmed(sochain_url, network, tx, True) for tx in txs])

    return confirmed_UTXOs

async def get_is_tx_confirmed(sochain_url, network, tx, return_if_is_confirmed):
    """Get is tx confirmed
    https://sochain.com/api/#get-is-tx-confirmed

    :param sochain_url: sochain url
    :type sochain_url: str
    :param network: mainnet or testnet
    :type network: str
    :param tx: transaction object
    :type tx: transaction object
    :returns: Confirmation object
    """
    try:
        api_url = f'{sochain_url}/is_tx_confirmed/{to_sochain_network(network)}/{tx["txid"]}'

        client = http3.AsyncClient(timeout=5)
        response = await client.get(api_url)

        if response.status_code == 200:
            response = json.loads(response.content.decode('utf-8'))['data']
            if return_if_is_confirmed:
                if response['is_confirmed']:
                    return tx
            return tx
        else:
            raise Exception(str(response.content.decode('utf-8')))
    except Exception as err:
        raise Exception(str(err))

async def broadcast_tx(sochain_url, network, tx_hex):
    """Broadcast transaction
    https://sochain.com/api#send-transaction

    :param sochain_url: sochain url
    :type sochain_url: str
    :param network: testnet or mainnet
    :type network: str
    :param tx_hex: tranaction hex
    :type tx_hex: str
    :returns: Transaction ID
    """
    try:
        api_url = f'{sochain_url}/send_tx/{to_sochain_network(network)}'

        client = http3.AsyncClient(timeout=5)
        response = await client.post(url=api_url, data={'tx_hex': tx_hex})

        if response.status_code == 200:
            res = json.loads(response.content.decode('utf-8'))['data']
            return res['txid']
        else:
            return json.loads(response.content.decode('utf-8'))['data']
    except Exception as err:
        raise Exception(str(err))


from xchainpy_client.models.tx_types import TX, TxHistoryParams, TxPage
from xchainpy_client.utxo_client import UTXOClient
from xchainpy_util.chain import Chain
from . models.client_types import BitcoinTxParams, BitcoinClientParams
from . import crypto, sochain_api
from . import utils


class Client(UTXOClient):
    sochain_url = blockstream_url = ''

    def __init__(self, params:BitcoinClientParams):
        """
        :param params: params
        :type params: BitcoinClientParams
        """
        UTXOClient.__init__(self, Chain.Bitcoin, params)

        self.set_sochain_url(params.sochain_url)
        self.set_blockstream_url(params.blockstream_url)

    def set_sochain_url(self, sochain_url:str):
        """Set/Update the sochain url

        :param sochain_url: The new sochain url
        :type sochain_url: str
        """
        self.sochain_url = sochain_url

    def set_blockstream_url(self, blockstream_url:str):
        """Set/Update the blockstream url

        :param blockstream_url: The new blockstream url
        :type blockstream_url: str
        """
        self.blockstream_url = blockstream_url

    def get_explorer_url(self) -> str:
        """Get explorer url
        :returns: the explorer url based on the network
        """
        return 'https://blockstream.info/testnet' if self.network == 'testnet' else 'https://blockstream.info'

    def get_explorer_address_url(self, address:str) -> str:
        """Get the explorer url for the given address

        :param address: address
        :type address: str
        :returns: The explorer url for the given address based on the network
        """
        return f'{self.get_explorer_url()}/address/{address}'

    def get_explorer_tx_url(self, tx_id:str) -> str:
        """Get the explorer url for the given transaction id

        :param tx_id: tx_id
        :type tx_id: str
        :returns: The explorer url for the given transaction id based on the network
        """
        return f'{self.get_explorer_url()}/tx/{tx_id}'

    def get_address(self, index:int=0):
        """Get the current address

        :returns: the current address
        :raises: Raises if phrase has not been set before. A phrase is needed to create a wallet and to derive an address from it.
        """
        if index < 0:
            raise Exception('index must be greater than zero')
        
        if self.phrase:
            derivation_path = self.root_derivation_paths.testnet if self.get_network() == 'testnet' else self.root_derivation_paths.mainnet
            address = crypto.mnemonic_to_address(mnemonic=self.phrase, derivation_path=derivation_path + str(index), network=self.get_network())

            if not address:
                raise Exception('Address not defined')

            return address
        raise Exception('Phrase must be provided')

    def validate_address(self, network, address):
        """Validate the given address

        :param network: testnet or mainnet
        :type network: str
        :param address: address
        :type address: str
        :returns: True or False
        """
        return utils.validate_address(network, address)

    async def get_balance(self, address):
        """Get the BTC balance of a given address

        :param address: BTC address
        :type address: str
        :returns: The BTC balance of the address
        """
        balance = await utils.get_balance(self.sochain_url, address, self.get_network())
        return balance

    async def get_transactions(self, params:TxHistoryParams) -> TxPage:
        """Get transaction history of a given address with pagination options
        By default it will return the transaction history of the current wallet

        :param params: params
        :type params: TxHistoryParams
        :returns: The transaction history
        """
        try:
            transactions = await sochain_api.get_transactions(self.sochain_url, self.get_network(), params.address)
            total = transactions['total_txs']
            offset = params.offset or 0
            limit = params.limit or 10
            transactions['txs'] = transactions['txs'][offset:offset+limit]
            txs = []
            for tx in transactions['txs']:
                tx = await sochain_api.get_tx(self.sochain_url, self.get_network(), tx['txid'])
                tx = utils.parse_tx(tx)
                txs.append(tx)
 
            return TxPage(total=total, txs=txs)

        except Exception as err:
            raise Exception(str(err))

    async def get_transaction_data(self, tx_id:str) -> TX:
        """Get the transaction details of a given transaction id

        if you want to give a hash that is for mainnet and the current self.network is 'testnet',
        you should call self.set_network('mainnet') (and vice versa) and then call this method.

        :param tx_id: The transaction id
        :type tx_id: str
        :returns: The transaction details of the given transaction id
        """
        try:
            tx = await sochain_api.get_tx(self.sochain_url, self.get_network(), tx_id)
            tx = utils.parse_tx(tx)
            return tx
        except Exception as err:
            raise Exception(str(err))

    async def _get_suggested_fee_rate(self):
        fee_rate = await sochain_api.get_suggested_tx_fee()
        return fee_rate

    def _calc_fee(self, fee_rate, memo):
        fee = utils.calc_fee(fee_rate, memo)
        return fee

    async def transfer(self, params:BitcoinTxParams):
        """Transfer BTC

        :param params: transfer params
        :type params: BitcoinTxParams
        :returns: the transaction hash
        """
        if not params.fee_rate:
            fee_rates = await self.get_fee_rates()
            params.fee_rate = fee_rates.fast

        spend_pending_UTXO = False if params.memo else True

        t, utxos = await utils.build_tx(sochain_url=self.sochain_url, amount=round(params.amount*10**8), recipient=params.recipient,
                                        memo=params.memo, fee_rate=params.fee_rate, sender=self.get_address(), network=self.get_network(), 
                                        spend_pending_UTXO=spend_pending_UTXO)

        derivation_path = self.root_derivation_paths.testnet if self.get_network() == 'testnet' else self.root_derivation_paths.mainnet
        derivation_path += str(params.wallet_index)
        private_key = crypto.mnemonic_to_private_key(self.phrase, derivation_path, self.get_network())
        private_bytes = private_key.Raw().ToBytes()
        t.sign(private_bytes)

        return await utils.broadcast_tx(self.blockstream_url, self.get_network(), t.raw_hex())



from xchainpy_client.models.tx_types import TxParams
from xchainpy_client.models.types import Network, RootDerivationPaths, XChainClientParams
from xchainpy_util.asset import Asset, AssetBTC


class BitcoinClientParams(XChainClientParams):

    def __init__(self, network:Network=Network.Testnet, phrase=None,
                 root_derivation_paths:RootDerivationPaths=RootDerivationPaths(
                     # note this isn't bip44 compliant, but it keeps the wallets generated compatible to pre HD wallets
                     mainnet="m/84'/0'/0'/0/",
                     testnet="m/84'/1'/0'/0/"),
                 sochain_url:str='https://sochain.com/api/v2', blockstream_url:str='https://blockstream.info'):
        """
        :param network: network
        :type network: str
        :param phrase: phrase 
        :type phrase: str
        :param root_derivation_paths: root_derivation_paths 
        :type root_derivation_paths: RootDerivationPaths
        :param sochain_url: sochainUrl 
        :type sochain_url: str
        :param blockstream_url: blockstreamUrl 
        :type blockstream_url: str
        """
        super().__init__(network, phrase, root_derivation_paths)
        self._sochain_url = sochain_url
        self._blockstream_url = blockstream_url

    @property
    def sochain_url(self):
        return self._sochain_url

    @sochain_url.setter
    def sochain_url(self, sochain_url):
        self._sochain_url = sochain_url

    @property
    def blockstream_url(self):
        return self._blockstream_url

    @blockstream_url.setter
    def blockstream_url(self, blockstream_url):
        self._blockstream_url = blockstream_url


class BitcoinTxParams(TxParams):
    def __init__(self, amount, recipient, memo='', fee_rate=None, wallet_index=0, asset:Asset=AssetBTC):
        """Transfer BTC

        :param amount: amount of BTC to transfer (don't multiply by 10**8)
        :type amount: int, float, decimal
        :param recipient: destination address
        :type recipient: str
        :param memo: optional memo for transaction
        :type memo: str
        :param fee_rate: fee rates for transaction
        :type fee_rate: int
        :param wallet_index: wallet_index
        :type wallet_index: int
        """
        TxParams.__init__(self, AssetBTC, amount, recipient, memo, wallet_index)
        self._fee_rate = fee_rate

    @property
    def fee_rate(self):
        return self._fee_rate

    @fee_rate.setter
    def fee_rate(self, fee_rate):
        self._fee_rate = fee_rate


